from demutil.util import vector_to_radian
import cairocffi as cairo
import math
import io


__alpha__ = 0.4


__COLORS__ = [
    (33/255, 93/255, 60/255, __alpha__),
    (29/255, 164/255, 79/255, __alpha__),
    (71/255, 208/255, 69/255, __alpha__),
    (155/255, 231/255, 65/255, __alpha__),
    (245/255, 216/255, 48/255, __alpha__),
    (245/255, 162/255, 48/255, __alpha__),
    (245/255, 94/255, 48/255, __alpha__),
    (245/255, 48/255, 48/255, __alpha__)
]


def draw_polygon(context, shape):
    path = shape.exterior
    context.move_to(path.coords[0][0], path.coords[0][1])
    for pt in path.coords[1:]:
        context.line_to(pt[0], pt[1])

    for path in shape.interiors:
        context.move_to(path.coords[0][0], path.coords[0][1])
        for pt in path.coords[1:]:
            context.line_to(pt[0], pt[1])

    context.close_path()


def get_image(shape, vectors):
    bound = shape.bounds
    start = [bound[0], bound[1]]
    width = bound[2] - bound[0]
    height = bound[3] - bound[1]

    # Create Image Surface
    svg = io.BytesIO()
    surface = cairo.SVGSurface(svg, width, height)
    # surface = cairo.SVGSurface("test.svg", width, height)
    context = cairo.Context(surface)

    # 상하 flip
    context.translate(0, height)
    context.scale(1, -1)

    # origin point 설정
    context.translate(-start[0], -start[1])

    context.set_source_rgb(1.0, 1.0, 1.0)
    context.paint()

    # box 그리기
    for pos in vectors:
        x = pos[0]
        y = pos[1]

        slope = math.degrees(vector_to_radian(pos[3:6]))
        idx = ((slope - 10.0) / 5.0)
        idx = 0 if idx < 0 else 7 if idx > 7 else idx
        context.set_source_rgba(*__COLORS__[int(idx)])
        context.rectangle(x, y, 10, 10)
        context.fill()

    # 윤곽선 그리기
    draw_polygon(context, shape)

    context.set_line_width(2.0)
    context.set_source_rgb(0.0, 123/255, 1.0)
    context.stroke_preserve()

    context.set_source_rgba(0.0, 0.0, 0.0, 0.8)
    context.set_operator(cairo.OPERATOR_OVERLAY)
    context.fill()

    # surface 에 그리기
    surface.finish()

    # optimization
    # print("Start optimization...")
    # options = scour.sanitizeOptions(options=None)
    # options.remove_metadata = True
    # options.remove_descriptive_elements = True
    # options.strip_comments = True
    # options.enable_viewboxing = True
    # options.indent_type = None
    # options.newlines = False
    # options.strip_ids = True
    # options.shorten_ids = True
    # options.strip_xml_prolog = True
    # options.digits = 3

    source = svg.getvalue().decode()
    # output = scour.scourString(source, options)

    print(len(source))

    return source



from .processor import DEMObject


class DEMLoader:
    def query(self, shape) -> DEMObject:
        pass

    def process(self):
        pass

    @classmethod
    class LoadFailedException(Exception):
        pass



import numpy as np
import math
from typing import Tuple, Optional


_t_Point = Optional[Tuple[float, float]]


class Envelope:
    _east: float = None
    _west: float = None
    _north: float = None
    _south: float = None

    def __init__(self, nw: _t_Point=None, ne: _t_Point=None, sw: _t_Point=None, se: _t_Point=None, from_bound=None, from_shape=None):
        if from_bound is not None:
            self._west = from_bound[0]
            self._east = from_bound[2]
            self._south = from_bound[1]
            self._north = from_bound[3]

        elif from_shape is not None:
            bound = from_shape.bounds

            self._west = bound[0]
            self._east = bound[2]
            self._south = bound[1]
            self._north = bound[3]

        elif nw is not None:
            self._north = nw[1]
            self._west = nw[0]

            if se is not None:
                self._south = se[1]
                self._east = se[0]
            else:
                raise TypeError("must have south-east point")
        elif ne is not None:
            self._north = ne[1]
            self._east = ne[0]

            if sw is not None:
                self._south = sw[1]
                self._west = sw[0]
            else:
                raise TypeError("must have south-west point")
        else:
            raise TypeError("must have north-east or north-west point")

    def west(self) -> float:
        return self._west

    def east(self) -> float:
        return self._east

    def north(self) -> float:
        return self._north

    def south(self) -> float:
        return self._south


def norm_array(arr):
    for vec in arr:
        yield vec / np.linalg.norm(vec)


def vector_to_radian(vec):
    ref = np.asarray([0, 0, 1])

    return math.acos(np.dot(vec, ref) / np.linalg.norm(vec) * np.linalg.norm(ref))


def vector_to_bearing(vec):
    bearing = math.atan2(vec[0], vec[1])
    bearing = (bearing + (2*math.pi)) % (2*math.pi)

    return math.degrees(bearing)


def bearing_to_repr(bearing):
    val = int((bearing/22.5)+.5)
    arr = ["N", "NNE", "NE", "ENE", "E", "ESE", "SE", "SSE", "S", "SSW", "SW", "WSW", "W", "WNW", "NW", "NNW"]

    return arr[(val % 16)]



import numpy as np
import itertools
from scipy.interpolate import griddata

from .util import Envelope, norm_array

from typing import List, Tuple


_t_DEM = List[Tuple[float, float, float]]


class DEMObject:
    _dem: _t_DEM = None

    def __init__(self, dem: _t_DEM):
        self._dem = np.asarray(dem)

    def __repr__(self):
        return f"<DEMObject @ {len(self._dem)}>"

    def _axis(self, bound: Envelope, resolution: float):
        sx, sy = bound.west(), bound.south()
        ex, ey = bound.east(), bound.north()

        x_axis = np.arange(sx, ex+resolution, resolution)
        y_axis = np.arange(sy, ey+resolution, resolution)

        return x_axis, y_axis

    # 특정 해상도 (meter) 단위로 elevation grid 생성
    def grid(self, bound: Envelope, resolution: float=10.0):
        DEM = self._dem

        x_axis, y_axis = self._axis(bound, resolution)

        X, Y = np.meshgrid(x_axis, y_axis)
        Q = np.vstack([X.ravel(), Y.ravel()]).T
        # print(Q)

        Z = griddata((DEM[:, 0], DEM[:, 1]), DEM[:, 2], (Q[:, 0], Q[:, 1]), method='linear', fill_value=0)
        return np.vstack([Q[:, 0], Q[:, 1], Z]).T

    def slope(self, bound: Envelope, resolution: float=10.0):
        x_axis, y_axis = self._axis(bound, resolution)
        grid = self.grid(bound, resolution)

        height = grid[:,2]
        height = height.reshape((len(y_axis), len(x_axis)))

        grad_y, grad_x = np.gradient(height)

        vec_x = np.asarray([[resolution, 0.0, x] for x in list(itertools.chain(*grad_x))])
        vec_x = np.asarray(list(norm_array(vec_x)))
        vec_y = np.asarray([[0.0, resolution, y] for y in list(itertools.chain(*grad_y))])
        vec_y = np.asarray(list(norm_array(vec_y)))

        slope = np.cross(vec_x, vec_y)
        slope = np.asarray(list(norm_array(slope)))

        return np.hstack([grid, slope])



from setuptools import setup, find_packages


setup(
    name='dem-util',
    version='1.0.3',
    url="https://github.com/BetaS/pydemutil",
    author="BetaS",
    author_email="thou1999@gmail.com",
    description="DEM slope modeling and imaging with trilinear interpolation",
    packages=find_packages(),
    install_requires=[
        'numpy',
        'scipy',
        'pyproj',
        'shapely',
        'cairocffi'
    ],
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    classifiers=[
        'Programming Language :: Python :: 3',
            "License :: OSI Approved :: MIT License"
    ]
)